from dataclasses import dataclass

from wth.utils.dataclass import deserialize


@dataclass
class B:
    name: str


@dataclass
class A:
    test_name: str = ""
    b: list[B] = None


t1 = deserialize([{"test-name": "wttch", "b": [{"name": "wb"}]}], list[A])
print(t1)
t1 = deserialize({"test-name": "wttch"}, A)
print(t1)



import random
from typing import Union

import numpy as np

#
# def seq_data_iter_random(corpus: np.ndarray, batch_size: int, seq_len: int, label=False) \
#         -> Union[tuple[np.ndarray, np.ndarray], np.ndarray]:
#     """
#     随机生成序列数据，只保证序列内的顺序和原始顺序一致，不能保证小批量数据之间连续。
#     Args:
#         corpus: 原始语料库
#         batch_size: 批处理大小
#         seq_len: 序列长度
#
#     Returns:
#
#     """
#     # 偏移
#     corpus = corpus[random.randint(0, seq_len):]
#     # -1, 因为标签 +1, 序列的下一个数据就是标签
#     num_seqs = (len(corpus) - (1 if label else 0)) // seq_len
#     # 抽样开始的位置
#     init_indices = list(range(0, num_seqs * seq_len, seq_len))
#     # 打乱, batch 内序列不一定连续
#     random.shuffle(init_indices)
#
#     def get_data(pos):
#         return corpus[pos: pos + seq_len]
#
#     # 可以生成多少批数据
#     num_batches = num_seqs // batch_size
#     for i in range(0, batch_size * num_batches, batch_size):
#         init_indices_per_batch = init_indices[i:i + batch_size]
#         X = [get_data(j) for j in init_indices_per_batch]
#         if label:
#             Y = [get_data(j + 1) for j in init_indices_per_batch]
#             yield np.array(X), np.array(Y)
#         else:
#             yield np.array(X)
#
#
# def seq_data_iter_sequential(corpus: np.ndarray, batch_size: int, seq_len: int, label: bool = False) \
#         -> Union[tuple[np.ndarray, np.ndarray], np.ndarray]:
#     """
#     顺序生成序列数据，能保证小批量数据之间连续。
#     Args:
#         corpus: 原始语料库
#         batch_size: 批处理大小
#         seq_len: 序列长度
#
#     Returns:
#
#     """
#     corpus = corpus[random.randint(0, seq_len):]
#     num_tokens = ((len(corpus) - (1 if label else 0)) // batch_size) * batch_size
#
#     Xs = corpus[0: num_tokens]
#     Xs = Xs.reshape(batch_size, -1)
#     if label:
#         Ys = corpus[1: num_tokens + 1]
#         Ys = Ys.reshape(batch_size, -1)
#
#     num_batches = Xs.shape[1] // seq_len
#     for i in range(0, seq_len * num_batches, seq_len):
#         X = Xs[:, i: i + seq_len]
#         if label:
#             Y = Ys[:, i: i + seq_len]
#             yield X, Y
#         else:
#             yield X
#
#
# class SeqDataGenerator:
#
#     def __init__(self, data: np.ndarray, batch_size: int, seq_len: int, shuffle: bool = True, label: bool = False):
#         """
#         序列数据生成器。
#         Args:
#             data: 原始数据
#             batch_size: 批处理长度
#             seq_len: 序列长度
#             shuffle: 批之间收否随机
#         """
#         # 分开处理是因为使用 yield, 也可以同一个函数处理，之后随机一下
#         self.data_iter_fn = seq_data_iter_random if shuffle else seq_data_iter_sequential
#
#         self.data = data
#         self.batch_size = batch_size
#         self.seq_len = seq_len
#         self.label = label
#
#     def __iter__(self):
#         return self.data_iter_fn(self.data, self.batch_size, self.seq_len, self.label)


class SeqDataLoader:
    def __init__(self, data: np.ndarray, seq_len: int, label: bool = False):
        """
        序列数据加载器。
        Args:
            data: 原始数据
            seq_len: 序列长度
            label: 是否生成 label: 对 x 向后偏移一位, 即预测的下一个
        """
        self.data = data
        self.seq_len = seq_len
        self.label = label

    def get_data(self):
        data = self.data[random.randint(0, self.seq_len):]

        num_seqs = (len(data) - (1 if self.label else 0)) // self.seq_len * self.seq_len

        x_data = data[0:num_seqs]
        y_data = data[1: num_seqs + 1]

        if self.label:
            return x_data.reshape((-1, self.seq_len,) + x_data.shape[1:]), y_data.reshape(
                (-1, self.seq_len,) + x_data.shape[1:])

        return x_data.reshape((-1, self.seq_len,) + x_data.shape[1:])



import logging
import os
import time

log_format_str = '%(asctime)s.%(msecs)03d %(levelname)s  [%(threadName)s] %(name)s - %(message)s'
colored_log_formatter = None
try:
    # 尝试使用 colorlog
    import colorlog

    colored_log_formatter = colorlog.ColoredFormatter(
        f'%(log_color)s{log_format_str}',
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'bold_red'
        },
        reset=True,
        style='%'
    )
except ImportError:
    print("⚠️:如果想要带颜色的日志，请安装 colorlog 库")


def get_logger(name: str = "") -> logging.Logger:
    """
    根据指定的名称生成一个 logger, 并尝试使用 colorlog.
    Args:
        name: 日记名称

    Returns:
        logging.Logger: 生成的日志
    """
    timestamp = int(time.time())

    timestamp_str = time.strftime("%Y%m%d-%H%M%S", time.localtime(timestamp))

    if not os.path.exists('logs'):
        os.makedirs('logs')

    log_formatter = logging.Formatter(log_format_str)

    custom_logger = logging.getLogger(name)
    custom_logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(f'logs/{timestamp_str}.log')
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(log_formatter)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.DEBUG)
    console_handler.setFormatter(log_formatter if colored_log_formatter is None else colored_log_formatter)

    custom_logger.addHandler(file_handler)
    custom_logger.addHandler(console_handler)

    return custom_logger



from typing import Any

import importlib


class ReflectUtil:

    @staticmethod
    def try_init_class(params: dict) -> Any:
        """
        尝试通过给定的模块位置, 类名, 初始化构造参数来生成对象。

        例如:
        {
            'module': 'wth.utils.reflect',
            'class': 'ReflectUtil',
            'init-kwargs': {}
        }

        `module` 指定要获取的模块位置,
        `class` 要获取的类名字
        `init-kwargs` 初始化构造参数, 该参数一定要和 指定的类的构造函数参数一致, 否则会初始化失败.
        """
        module_name = params['module']
        class_name = params['class']
        init_args = params['init-kwargs'] if 'init-kwargs' in params else {}

        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return class_(**init_args)



from torch.utils.data import Dataset, random_split, DataLoader


class TrainAndTestDataLoader:
    def __init__(self, dataset: Dataset, train_rate: float = 0.9, config=None):
        self.dataset = dataset
        if train_rate < 0 or train_rate > 1:
            raise ValueError(f"train_rate should be between 0 and 1, got {train_rate}")
        batch_size = config.batch_size if config is not None else 64

        train_dataset, test_dataset = random_split(self.dataset, lengths=[train_rate, 1 - train_rate])
        self.train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=4)
        self.test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=True, num_workers=4)



from abc import abstractmethod

from .base import BaseNotification


class App:
    """
    为了发送异常信息。
    """

    def __init__(self, notification: BaseNotification, name: str = ""):
        self.notification = notification
        self.name = ""

    def start(self):
        try:
            self._run()
        except Exception as e:
            self.notification.send_text(f"应用[{self.name}]出现异常: {str(e)}")
            raise e

    @abstractmethod
    def _run(self):
        pass



from dataclasses import is_dataclass, fields
from types import GenericAlias


def _from_dict(data, cls):
    if isinstance(data, list):
        # 如果是列表, 循环进行转换
        elem_type = cls.__args__[0]
        return [_from_dict(item, elem_type) for item in data]
    if not is_dataclass(cls):
        # 判断是否是 dataclass
        return data
    if isinstance(data, dict):
        # 如果是字典的, 检查各个字段递归进行转换
        field_types = {_handling_identifier(f.name): f.type for f in fields(cls)}
        return cls(**{_handling_identifier(f): _from_dict(data[f], field_types[_handling_identifier(f)]) for f in data})
    else:
        return data


def _handling_identifier(identifier: str):
    # 将横线转换为下划线
    # yml 中定义一般使用的是横线, 但是变量标识符不可以
    return identifier.replace("-", "_")


def deserialize(data, cls):
    """
    反序列化 dataclass
    Args:
        data: 数据字典，json loads 出来的
        cls: 要转的 dataclass 类名称
    """
    return _from_dict(data, cls)



from distutils.core import setup

setup(
    name='wttch-train-helper',
    version='0.0.39',
    description="Wttch's Train Helper",
    author="wttch",
    packages=[
        'wth',
        'wth.data',
        'wth.notification',
        'wth.torch',
        'wth.torch.utils',
        'wth.utils',
    ],
    package_dir={
        '': 'src/wttch-train-helper'
    },
    requires=['torch', 'numpy', 'prettytable']
)



import json

import yaml

from wth.log import get_logger


class Config:
    """
    环境配置，使用 yml 文件。
    """

    def __init__(self, key: str, config_file='config.yml'):
        """
        构造函数。
        Args:
            config_file: 配置文件所在位置，默认 config.yml。
        """
        self.config_file = config_file
        self.logger = get_logger("Config")
        self.key = key
        with open(config_file) as f:
            self._original = yaml.load_all(f, Loader=yaml.FullLoader)
            self._active_key = None
            self._shared = {}

            config_map = {}
            for config in self._original:
                config = config['config']
                # 获取激活的环境
                if 'active' in config:
                    self._active_key = config['active']
                    continue
                # 共享的数据
                if 'shared' in config and config['shared']:
                    self._shared = config
                    continue

                if 'name' in config:
                    name = config['name']
                    config_map[name] = config
                else:
                    raise ValueError(f"配置中存在没有 name 的配置. {config}")
            if self._active_key is None:
                raise ValueError("No active config.")
            # 激活的环境
            self._active = config_map[self._active_key]
            if self._active is None:
                raise ValueError("Active config not found.")

            self.logger.info(f"Active config env[{self._active_key}].")
            if len(self._shared) != 0:
                self.logger.info(f"存在共享配置.")

    def __getitem__(self, *items):
        try:
            value = self._active[self.key]
            for item in items:
                value = value[item]
            return value
        except (KeyError, ValueError):
            pass
        # 尝试共享获取
        try:
            value = self._shared[self.key]
            for item in items:
                value = value[item]
            return value
        except (KeyError, ValueError):
            pass
        return None

    def __str__(self):
        return json.dumps({
            "active": self._active.get(self.key, {}),
            "shared": self._shared.get(self.key, {}),
        })
